<?php
// This file is part of Moodle - http://moodle.org/
// Moodle is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License.

/**
 * Library file for the Zoom Online block.
 *
 * This file contains the pluginfile function for handling file access.
 *
 * @package   block_zoomonline
 * @copyright 2024 Your Name
 * @license   http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

/**
 * Handles file access for the zoomonline block.
 *
 * @param stdClass $course The course object.
 * @param stdClass $birecord_or_cm The block instance or course module object.
 * @param context $context The context object.
 * @param string $filearea The name of the file area.
 * @param array $args Additional arguments (itemid, path).
 * @param bool $forcedownload Whether or not to force download.
 * @param array $options Additional options affecting file serving.
 * @return void
 */
function block_zoomonline_pluginfile($course, $birecord_or_cm, $context, $filearea, $args, $forcedownload, array $options = array()) {
    global $CFG;

    // Debugging message to confirm the function is called
    debugging("block_zoomonline_pluginfile was called", DEBUG_DEVELOPER);

    // Validate the context is block level
    if ($context->contextlevel != CONTEXT_BLOCK) {
        send_file_not_found();
    }

    // Validate the file area
    if ($filearea !== 'zoom_videos') {
        send_file_not_found();
    }

    // Ensure the user has logged in and has the proper capability
    require_login($course);
 

    // Parse arguments to get itemid, filepath, and filename
    $itemid = array_shift($args);
    $filename = array_pop($args);
    $filepath = '/' . implode('/', $args) . '/';

    // Use Moodle's file storage API to fetch the file
    $fs = get_file_storage();
    $file = $fs->get_file($context->id, 'block_zoomonline', $filearea, $itemid, $filepath, $filename);

    // Check if the file exists and is not a directory
    if (!$file || $file->is_directory()) {
        send_file_not_found();
    }

    // Serve the file to the user
    send_stored_file($file, 0, 0, $forcedownload, $options);
}

